/***************************************************************************
 BioSequence.h  -  description
 -------------------
 begin                : Thu Jun 16 2002
 copyright            : (C) 2002 by Oleksiy Kohany
 email                : kohany@girinst.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef _BIO_SEQUENCE_H_
#define _BIO_SEQUENCE_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <list>
#include <vector>
#include <iostream>
#include <stdio.h>
#include <stdlib.h>

using namespace std;
namespace bio {

#define GAPSYMB '-'
#define STOPCODONSYMB '*'

#define LINELENGTH 70

#define BASERANGE 128

   enum SEQTYPE {
      SEQTYPE_UNKNOWN, NA, AA
   };

   //const char sValidASCII[] = "ABCDEFGHIKLMNPQRSTUVWXYZabcdefghiklmnpqrstuvwxyz";

   class BioFragment: public pair<int, int> {
   public:
      string locusName;
      BioFragment(string name, int p1, int p2) :
         pair<int, int> (p1, p2), locusName(name) {
      }
      BioFragment(const char* name, int p1, int p2) :
         pair<int, int> (p1, p2), locusName(name) {
      }
      size_t length() const {
         return abs(second - first) + 1;
      }
   };

   /**
    * Generally it is not a good practice to create subclasses of string
    * as well as subclasses of other STL containers. Since string does not
    * have a virtual destructor, a care must be taken NOT to delete any
    * subclasses of string by pointer to base string object.
    */
   class BioSequence: public string {
   protected:
      string m_sName;
      mutable SEQTYPE seqtype;
      /** ---------------------------------------------------------- */
      void rmTrash();
   public:
      using string::assign;

      BioSequence& operator =(const string& s);
      BioSequence& operator =(const char* s);

      BioSequence& append(const BioSequence& s);
      BioSequence& append(const string& s);
      BioSequence& append(const char* s);
      BioSequence& append(char c);

      BioSequence& assign(const char* s);
      BioSequence& assign(const char* s, size_type n);
      BioSequence& assign(const string& s);
      BioSequence& assign(const string &sSeq, const string &sName);
      BioSequence& assign(const char *pcSeq, const char *pcName);
      /** ---------------------------------------------------------- */
      virtual BioSequence& assign(const BioSequence &Seq);
      /** No descriptions */
      virtual int readHead(istream& is);
      /** ---------------------------------------------------------- */
      void formatedPrint(ostream &os, size_t iLineLen = LINELENGTH) const;
      /** ---------------------------------------------------------- */
      void unformatedPrint(ostream &os) const;
      virtual void clear();
      int removeGaps();
      /** ---------------------------------------------------------- */
      BioSequence();
      /** ---------------------------------------------------------- */
      BioSequence(const BioSequence& other);
      /** ---------------------------------------------------------- */
      BioSequence(const string &sSeq, const string &sName);
      /** ---------------------------------------------------------- */
      BioSequence(const char *pcSeq, const char *pcName);
      /** ---------------------------------------------------------- */
      virtual ~BioSequence();
      /** ---------------------------------------------------------- */
      const string& getName() const;
      /** ---------------------------------------------------------- */
      void setName(const string &sName);
      /** ---------------------------------------------------------- */
      void setName(const char *pcName);
      /** ---------------------------------------------------------- */
      static string& complement(string& str);
      /** ---------------------------------------------------------- */
      virtual BioSequence& complement();
      /** ---------------------------------------------------------- */
      BioSequence& operator =(const BioSequence &Seq);
      /** ---------------------------------------------------------- */
      long findBase(long iPos = 0) const;
      /** ---------------------------------------------------------- */
      BioSequence& format();
      /** ---------------------------------------------------------- */
      BioSequence& sqeezeOutGaps();
      /** ---------------------------------------------------------- */
      virtual void printHead(ostream &os) const;
      /** ---------------------------------------------------------- */
      virtual void printFASTAHead(ostream &os) const;
      /** ---------------------------------------------------------- */
      virtual void printIGHead(ostream &os) const;
      /** ---------------------------------------------------------- */
      virtual void print(ostream& os, bool format = true) const;
      /** ---------------------------------------------------------- */
      virtual void printFASTA(ostream& os, bool format = true) const;
      /** ---------------------------------------------------------- */
      virtual void printIG(ostream& os, bool format = true) const;
      /** ---------------------------------------------------------- */
      static char complement(char c);
      /** ---------------------------------------------------------- */
      virtual int read(istream& is) throw ();
      /** ---------------------------------------------------------- */
      BioSequence& toupper();
      /** ---------------------------------------------------------- */
      BioSequence& tolower();
      /** No descriptions */
      SEQTYPE guessType() const;
      /** No descriptions */
      /** No descriptions */
      size_t effective_length();
      /** No descriptions */
      int baseCount2position(int n);
      /** Read property of SEQTYPE seqtype. */
      virtual SEQTYPE getseqtype() const;
      static string& shrink(string& s, char c);
      const BioSequence subseq(size_t pos, size_t n) const;
      static const string subseq(const string& s, size_t pos, size_t n);
      static vector<char> string2vector(const string& s);
      static string vector2string(const vector<char>& v);
      string::const_iterator next_base(string::const_iterator ptr, int* gaplen = NULL) const;
      string::const_iterator prev_base(string::const_iterator ptr, int* gaplen = NULL) const;
      char next_base(int p, int* gaplen = NULL) const;
      char prev_base(int p, int* gaplen = NULL) const;
      static int baseCount2position(const string& s, int n);
      size_t baseCount(size_t pos, int n) const;
      static size_t baseCount(const string& s, size_t pos, int n);
      void replaceChar(char target, char replacement);
   };
   /** ========================================================== */
   istream& operator >>(istream &os, BioSequence& seq);
   ostream& operator <<(ostream &os, const BioSequence& seq);

   /** ========================================================== */
   class BioSubSequence: public BioSequence {
   protected:
      //size_t m_iFrom;
      //size_t m_iTo;
      string m_sComment;
      list<BioFragment> fragments;
   public:
      using BioSequence::assign;
      using BioSequence::append;
      using BioSequence::operator =;
      BioSubSequence();
      BioSubSequence(const BioSequence& other);
      BioSubSequence(const BioSubSequence& other);
      BioSubSequence(const string &sSeq, const string &sName, size_t iFrom, size_t iTo);
      BioSubSequence(const char *pcSeq, const char *pcName, size_t iFrom, size_t iTo);
      size_t getFrom() const;
      size_t getTo() const;
      const string& getComment() const;
      void setFrom(size_t iFrom);
      void setTo(size_t iTo);
      void addComment(const string& sComment);
      void setComment(const string& sComment);
      void setComment(const char *pcComment);
      BioSubSequence& complement();
      virtual void printHead(ostream &os) const;
      virtual void printIGHead(ostream &os) const;
      virtual void printFASTAHead(ostream &os) const;
      virtual BioSubSequence& assign(const BioSequence& sSeq);
      virtual BioSubSequence& assign(const BioSubSequence& sSeq);
      BioSequence& trunc_left(size_t pos);
      BioSequence& trunc_right(size_t pos);
      const BioSubSequence subseq(size_t pos, size_t n, bool literal = false) const;
      void adjust_position(int start, int end);
      virtual int readHead(istream& is);
      BioSubSequence& append(BioSubSequence& other);
      const list<BioFragment>& getFragments() const {
         return fragments;
      }
      list<BioFragment>& getFragments() {
         return fragments;
      }
      virtual void clear();
   };
   class RatedSequence: public BioSequence {
   protected:
      double score;
   public:
      using string::assign;
      RatedSequence();
      RatedSequence(const BioSequence& other);
      RatedSequence(const RatedSequence& other);
      virtual RatedSequence& assign(const RatedSequence& other);
      void setScore(double score);
      double getScore() const;
      virtual void clear();
      virtual void printHead(ostream &os) const;
      virtual void printIGHead(ostream &os) const;
      virtual void printFASTAHead(ostream &os) const;
   };

   template<class ForwardIterator>
   size_t effective_length(ForwardIterator beg, ForwardIterator end) {
      size_t iLen = 0;
      for (ForwardIterator sptr = beg; sptr != end; sptr++) {
         if (isalpha(*sptr)) {
            iLen++;
         }
      }
      return iLen;
   }
}
//==========================================================

#endif //_BIO_SEQUENCE_H_
