/*
 * ThreadedAligner.h
 *
 *  Created on: Dec 1, 2009
 *      Author: kohany
 */

#ifndef THREADEDALIGNER_H_
#define THREADEDALIGNER_H_

#include <assert.h>
#include <stdexcept>
#include "ThreadPool.h"
#include "BioAlignment.h"
#include "FarrarSWAligner.h"
#include "MonardoSWAligner.h"
namespace bio {

   class ThreadedAligner {
      /**
       * Stores synchronization primitives for multithreaded alignment functions
       * as well as thread termination flag and thread count for barrier initialization.
       */
      struct SyncData {
         int threadCount;
         pthread_mutex_t optr_mutex;
         pthread_mutex_t iptr_mutex;
         SyncData(int threadCount) :
            threadCount(threadCount) {
            if (pthread_mutex_init(&optr_mutex, NULL) != 0) {
               throw(runtime_error("Failed to initialize write_mutex"));
            }
            if (pthread_mutex_init(&iptr_mutex, NULL) != 0) {
               throw(runtime_error("Failed to initialize read_mutex"));
            }
         }
         virtual ~SyncData() {
            pthread_mutex_destroy(&optr_mutex);
            pthread_mutex_destroy(&iptr_mutex);
         }
      protected:
         SyncData() {
         }
      };

      /**
       * Stores input and output iterators shared between multiple threads
       * performing many-to-many type of alignments.
       */
      template<class ForwardIterator, class OutputIterator>
      struct ManyToMany: public SyncData {
         ManyToMany(ForwardIterator ib1, ForwardIterator ie1, ForwardIterator ib2, ForwardIterator ie2, OutputIterator oi,
                    const SWAligner& aligner, int threadCount) :
            SyncData(threadCount), score(0), ib1(ib1), ip1(ib1), ie1(ie1), ib2(ib2), ip2(ip2), ie2(ie2), oi(oi), aligner(aligner) {
         }
         ~ManyToMany() {
         }
         double score;
         ForwardIterator ib1; //input 1 begin pointer
         ForwardIterator ip1; //input 1 current iterator
         ForwardIterator ie1; //input 1 end pointer
         ForwardIterator ib2; //input 2 begin pointer
         ForwardIterator ip2; //input 2 current iterator
         ForwardIterator ie2; //input 2 end pointer
         OutputIterator oi; //output iterator
         const SWAligner& aligner;
      };

      /**
       * Stores input and output iterators shared between multiple threads
       * performing one-to-many type of alignments.
       */
      template<class ForwardIterator, class OutputIterator>
      struct OneToMany: public SyncData {
         typedef typename std::iterator_traits<ForwardIterator>::reference reference_t;
         double score;
         OneToMany(const reference_t ref, ForwardIterator ib, ForwardIterator ie, OutputIterator oi, const SWAligner& aligner,
                   int threadCount) :
            SyncData(threadCount), score(0), ref(ref), ib(ib), ip(ib), ie(ie), oi(oi), aligner(aligner) {
         }
         ~OneToMany() {
         }
         const reference_t ref;
         ForwardIterator ib;
         ForwardIterator ip;
         ForwardIterator ie;
         OutputIterator oi;
         const SWAligner& aligner;
      };

      ThreadPool threadPool;
      static bool isMultithreaded;
      auto_ptr<SWAligner> pAligner;
   public:
      ThreadedAligner();
      virtual ~ThreadedAligner();
      void setGapIni(int gapIni) {
         pAligner->setGapIniPenalty(gapIni);
      }
      void setGapExt(int gapExt) {
         pAligner->setGapExtPenalty(gapExt);
      }
      void setAligner(auto_ptr<SWAligner> aligner) {
         pAligner = aligner;
      }
      static void setMultithreaded(bool value);
      static bool getMultithreaded();
      void setAlignmentMatrix(const char* matfile) {
         pAligner->setAlignmentMatrix(matfile);
      }
      void setAlignmentMatrix(const SWAligner::ScoreMatrix_t& mat) {
         pAligner->setAlignmentMatrix(mat);
      }
      //=============================================================================
      template<class ForwardIterator, class OutputIterator>
      static void* alignOneToMany(void* arg);
      /**
       * =============================================================================
       * This method must be called from a single threaded code.
       * Multithreading is implemented inside this method.
       * =============================================================================
       */
      template<class ForwardIterator, class OutputIterator>
      double align(const typename std::iterator_traits<ForwardIterator>::reference ref, ForwardIterator beg, ForwardIterator end,
                   OutputIterator out);
      /**
       * Aligns all sequences from the first range to all sequences from the second range represented by
       * iterators in struct ManyToMany.
       */
      template<class ForwardIterator, class OutputIterator>
      static void* alignManyToMany(void* arg);
      /**
       * =============================================================================
       * Aligns all sequences in the first range to all sequences from the second.
       * This method must be called from a single threaded code.
       * Multithreading is implemented inside this method.
       * =============================================================================
       */
      template<class ForwardIterator, class OutputIterator>
      double align(ForwardIterator beg1, ForwardIterator end1, ForwardIterator beg2, ForwardIterator end2, OutputIterator out);
      /**
       * Aligns all sequences from the first range to all sequences from the second range represented by
       * iterators in struct ManyToMany.
       */
      template<class ForwardIterator, class OutputIterator>
      static void* alignCorresponding(void* arg);

      /**
       * =============================================================================
       * Aligns sequences from the first range to sequences from the second range that
       * are at corresponding positions - first to first, second to second,....
       * This method must be called from a single threaded code.
       * Multithreading is implemented inside this method.
       * =============================================================================
       */
      template<class ForwardIterator, class OutputIterator>
      double align_pairs(ForwardIterator beg1, ForwardIterator end1, ForwardIterator beg2, ForwardIterator end2, OutputIterator out);
      /**
       * =============================================================================
       * Aligns pairs of sequences read from input range:
       * first to second, third to fourth,...
       * This method must be called from a single threaded code.
       * Multithreading is implemented inside this method.
       * =============================================================================
       */
      template<class ForwardIterator, class OutputIterator>
      double align_pairs(ForwardIterator beg, ForwardIterator end, OutputIterator out);
   };

   template<class ForwardIterator, class OutputIterator>
   void* ThreadedAligner::alignOneToMany(void* arg) {
      OneToMany<ForwardIterator, OutputIterator>* args = (OneToMany<ForwardIterator, OutputIterator>*) arg;
      while ((args->threadCount > 1 && pthread_mutex_lock(&(args->iptr_mutex)) == 0 || args->threadCount <= 1) && args->ip != args->ie) {
         ForwardIterator sptr(args->ip);
         ++args->ip;
         if (args->threadCount > 1) {
            pthread_mutex_unlock(&(args->iptr_mutex));
         }
         /**
          * Do not align reference sequence to itself. This will work only if
          * reference sequence is within the range referenced by input iterator.
          */
         if (&(args->ref) == &(*sptr)) {
            //cerr << "Skipping alignment to itself " << args->ref.getName() << endl;
            continue;
         }
         if ((*sptr).size() == 0) {
            continue;
         }
         BioAlignment aln;
         aln.getCon().assign(args->ref);
         aln.getSeq().assign(*sptr);
         if (args->aligner.align(aln)) {
            if (args->threadCount > 1) {
               assert(pthread_mutex_lock(&args->optr_mutex) == 0);
            }
            *(args->oi) = aln;
            ++args->oi;
            args->score += aln.getScore();
            if (args->threadCount > 1) {
               pthread_mutex_unlock(&args->optr_mutex);
            }
         }
      }
      if (args->threadCount > 1) {
         pthread_mutex_unlock(&(args->iptr_mutex));
      }
   }

   template<class ForwardIterator, class OutputIterator>
   double ThreadedAligner::align(const typename std::iterator_traits<ForwardIterator>::reference ref, ForwardIterator beg,
                                 ForwardIterator end, OutputIterator out) {
      // Phill Green's implementation is not thread safe
      if (!isMultithreaded) {
         // If single threaded execution is needed
         OneToMany<ForwardIterator, OutputIterator> args(ref, beg, end, out, *pAligner, 1);
         alignOneToMany<ForwardIterator, OutputIterator> ((void*) &args);
         return args.score;
      }
      OneToMany<ForwardIterator, OutputIterator> args(ref, beg, end, out, *pAligner, threadPool.getThreadCount() + 1);
      threadPool.initWorkers(alignOneToMany<ForwardIterator, OutputIterator> , &args);
      threadPool.joinWorkers();
      return args.score;;
   }

   template<class ForwardIterator, class OutputIterator>
   void* ThreadedAligner::alignManyToMany(void* arg) {
      ManyToMany<ForwardIterator, OutputIterator>* args = (ManyToMany<ForwardIterator, OutputIterator>*) arg;
      while (args->ip1 != args->ie1) {
         while ((args->threadCount > 1 && pthread_mutex_lock(&(args->iptr_mutex)) == 0 || args->threadCount <= 1) && args->ip1 != args->ie1
               && args->ip2 != args->ie2) {
            ForwardIterator sptr1(args->ip1);
            ForwardIterator sptr2(args->ip2);
            ++args->ip2;
            if (args->threadCount > 1) {
               pthread_mutex_unlock(&(args->iptr_mutex));
            }
            /**
             * Do not align reference sequence to itself. This will work only if
             * reference sequence is within the range of input iterator.
             */
            if (&(*sptr1) == &(*sptr2)) {
               continue;
            }
            if ((*sptr2).size() == 0) {
               continue;
            }
            BioAlignment aln;
            aln.getCon().assign(*sptr1);
            aln.getSeq().assign(*sptr2);
            if (args->aligner.align(aln)) {
               if (args->threadCount > 1) {
                  assert(pthread_mutex_lock(&args->optr_mutex) == 0);
               }
               *(args->oi)++ = aln;
               args->score += aln.getScore();
               if (args->threadCount > 1) {
                  pthread_mutex_unlock(&args->optr_mutex);
               }
            }
         }
         /**
          * I the end of the second range is reached, rewind it to the beginning and increment
          * iterator on the first range.
          */
         if (args->ip2 == args->ie2) {
            args->ip2 = args->ib2;
            do {
               ++args->ip1;
            } while (args->ip1 != args->ie1 && *(args->ip1).size() == 0);
         }
         if (args->threadCount > 1) {
            pthread_mutex_unlock(&(args->iptr_mutex));
         }
      }
   }

   template<class ForwardIterator, class OutputIterator>
   double ThreadedAligner::align(ForwardIterator beg1, ForwardIterator end1, ForwardIterator beg2, ForwardIterator end2, OutputIterator out) {

      // Phill Green's implementation is not thread safe
      if (!isMultithreaded) {
         // If single threaded execution is needed
         ManyToMany<ForwardIterator, OutputIterator> args(beg1, end1, beg2, end2, out, *pAligner, 1);
         alignManyToMany<ForwardIterator, OutputIterator> ((void*) &args);
         return args.score;
      }
      ManyToMany<ForwardIterator, OutputIterator> args(beg1, end1, beg2, end2, out, *pAligner, threadPool.getThreadCount() + 1);
      initWorkers(alignManyToMany<ForwardIterator, OutputIterator> , &args);
      threadPool.joinWorkers();
      return args.score;
   }

   template<class ForwardIterator, class OutputIterator>
   void* ThreadedAligner::alignCorresponding(void* arg) {
      ManyToMany<ForwardIterator, OutputIterator>* args = (ManyToMany<ForwardIterator, OutputIterator>*) arg;
      while ((args->threadCount > 1 && pthread_mutex_lock(&(args->iptr_mutex)) == 0 || args->threadCount <= 1) && args->ip1 != args->ie1
            && args->ip2 != args->ie2) {
         ForwardIterator sptr1(args->ip1);
         ++args->ip1;
         ForwardIterator sptr2(args->ip2);
         ++args->ip2;
         if (args->threadCount > 1) {
            pthread_mutex_unlock(&(args->iptr_mutex));
         }
         if ((*sptr1).size() == 0) {
            continue;
         }
         if ((*sptr2).size() == 0) {
            continue;
         }
         BioAlignment aln;
         aln.getCon().assign(*sptr1);
         aln.getSeq().assign(*sptr2);
         if (args->aligner.align(aln)) {
            if (args->threadCount > 1) {
               assert(pthread_mutex_lock(&args->optr_mutex) == 0);
            }
            *(args->oi) = aln;
            ++args->oi;
            args->score += aln.getScore();
            if (args->threadCount > 1) {
               pthread_mutex_unlock(&args->optr_mutex);
            }
         }
      }
      if (args->threadCount > 1) {
         pthread_mutex_unlock(&(args->iptr_mutex));
      }
   }

   template<class ForwardIterator, class OutputIterator>
   double ThreadedAligner::align_pairs(ForwardIterator beg1, ForwardIterator end1, ForwardIterator beg2, ForwardIterator end2,
                                       OutputIterator out) {

      // Phill Green's implementation is not thread safe
      if (!isMultithreaded) {
         // If single threaded execution is needed
         ManyToMany<ForwardIterator, OutputIterator> args(beg1, end1, beg2, end2, out, *pAligner, 1);
         alignCorresponding<ForwardIterator, OutputIterator> ((void*) &args);
         return args.score;
      }
      ManyToMany<ForwardIterator, OutputIterator> args(beg1, end1, beg2, end2, out, *pAligner, threadPool.getThreadCount() + 1);
      initWorkers(alignCorresponding<ForwardIterator, OutputIterator> , &args);
      threadPool.joinWorkers();
      return args.score;
   }

   template<class ForwardIterator, class OutputIterator>
   double ThreadedAligner::align_pairs(ForwardIterator beg, ForwardIterator end, OutputIterator out) {

      // Phill Green's implementation is not thread safe
      if (!isMultithreaded) {
         // If single threaded execution is needed
         ManyToMany<ForwardIterator&, OutputIterator> args(beg, end, beg, end, out, *pAligner, 1);
         alignCorresponding<ForwardIterator, OutputIterator> ((void*) &args);
         return args.score;
      }
      ManyToMany<ForwardIterator&, OutputIterator> args(beg, end, beg, end, out, *pAligner, threadPool.getThreadCount() + 1);
      initWorkers(alignCorresponding<ForwardIterator, OutputIterator> , &args);
      threadPool.joinWorkers();
      return args.score;
   }
}
#endif /* THREADEDALIGNER_H_ */
