/*
 * FarrarSWAligner.cpp
 *
 *  Created on: Apr 6, 2010
 *      Author: Oleksiy Kohany <kohany@gmail.com>
 */
/******************************************************************
 *  Copyright 2006 by Michael Farrar.  All rights reserved.
 *  The following code may not be sold or incorporated into a commercial product,
 *  in whole or in part, without written consent of Michael Farrar.  For
 *  further information regarding permission for use or reproduction, please
 *  contact: Michael Farrar at farrar.michael@gmail.com.
 ********************************************************************/

/**
 *  Written by Michael Farrar, 2006.
 *  Modified by Oleksiy Kokhanyy (kohany@gmail.com), 2009 - added backtracking
 *  matrix in swStripedWord to produce alignment.
 *  Please send bug reports and/or suggestions to kohany@gmail.com
 **/
#include "FarrarSWAligner.h"

namespace bio {

   FarrarSWAligner::FarrarSWAligner() {
   }

   FarrarSWAligner::~FarrarSWAligner() {
   }

   auto_ptr<SwData> FarrarSWAligner::getProfile(const vindex_t& s) const {
      auto_ptr<SwFarrarData> pProfile(new SwFarrarData());
      size_t queryLength = s.size();
      pProfile->alloc(queryLength, scoreMatrix.getAlphabethSize());

      /* Fill in the short query profile */
      short* ps = (short *) pProfile->pvsQueryProf;
      int segSize = (queryLength + 7) / 8;
      int nCount = segSize * 8;
      for (int i = 0; i < scoreMatrix.getAlphabethSize(); ++i) {
         const score_t* matrixRow = scoreMatrix.getMatrix() + i * scoreMatrix.getAlphabethSize();
         for (int j = 0; j < segSize; ++j) {
            for (int k = j; k < nCount; k += segSize) {
               score_t weight;
               if (k >= queryLength) {
                  weight = 0;
               } else {
                  weight = matrixRow[s[k]];
               }
               *ps++ = (unsigned short) weight;
            }
         }
      }
      return auto_ptr<SwData> (pProfile);
   }

   /* ======================================================== */
   int FarrarSWAligner::swStripedWord(const vindex_t& topSeq, const vindex_t& botSeq, unsigned short gapOpen, unsigned short gapExtend,
                                      __m128i *pvQueryProf, __m128i *pvHLoad, __m128i *pvHStore, __m128i *pvE, __m128i *pvEGLLoad,
                                      __m128i *pvEGLStore, __m128i *pvDIR, BioSubSequence& top, BioSubSequence& bot) const {
      int i, j;

      int cmp;
      size_t topSeqLength = topSeq.size();
      size_t botSeqLength = botSeq.size();
      int iter = (topSeqLength + 7) / 8;

      __m128i *pv;

      __m128i vE, vF, vH;

      __m128i vMaxScore;
      __m128i vGapOpen;
      __m128i vGapExtend;
      __m128i vZERO;

      __m128i vMin;
      __m128i vMinimums;
      __m128i vTemp;

      __m128i *pvScore;

      /* Load gap opening penalty to all elements of a constant */
      vGapOpen = _mm_insert_epi16(vGapOpen, gapOpen, 0);
      vGapOpen = _mm_shufflelo_epi16(vGapOpen, 0);
      vGapOpen = _mm_shuffle_epi32(vGapOpen, 0);

      /* Load gap extension penalty to all elements of a constant */
      vGapExtend = _mm_insert_epi16(vGapExtend, gapExtend, 0);
      vGapExtend = _mm_shufflelo_epi16(vGapExtend, 0);
      vGapExtend = _mm_shuffle_epi32(vGapExtend, 0);

      /*  load vMaxScore with the zeros.  since we are using signed */
      /*  math, we will bias the maxscore to -32768 so we have the */
      /*  full range of the short. */
      vMaxScore = _mm_cmpeq_epi16(vMaxScore, vMaxScore);
      vMaxScore = _mm_slli_epi16(vMaxScore, 15);

      vMinimums = _mm_shuffle_epi32(vMaxScore, 0);

      vMin = _mm_shuffle_epi32(vMaxScore, 0);
      vMin = _mm_srli_si128(vMin, 14);

      vZERO = _mm_cmpgt_epi16(vZERO, vZERO);
      /* Zero out the storage vector */
      for (i = 0; i < iter; i++) {
         _mm_store_si128(pvE + i, vMaxScore);
         _mm_store_si128(pvHStore + i, vMaxScore);
         _mm_store_si128(pvEGLStore + i, vZERO);
         _mm_store_si128(pvDIR + i, vZERO);
      }

      vector<short> pathMatrix((botSeqLength + 1) * (topSeqLength + 1), 0);

      __m128i vMSK, vMSK2, vEGL, vFGL, vDIR;
      __m128i vONE = _mm_set1_epi16(1);

      //int iMaxScore_i = 0, iMaxScore_j = 0, iMaxScore_r = 0;
      __m128i vMaxScore_i, vMaxScore_j;
      vMaxScore_j = vMaxScore_i = _mm_set1_epi16(0);
      short iMaxScore = -0x8000;

      for (i = 0; i < botSeqLength; ++i) {
         /* fetch first data asap. */
         pvScore = pvQueryProf + botSeq[i] * iter;

         /* zero out F. */
         vF = _mm_cmpeq_epi16(vF, vF);
         vF = _mm_slli_epi16(vF, 15);
         /* zero out db gap lengths*/
         vFGL = _mm_cmpgt_epi16(vFGL, vFGL);

         /* load the next h value */
         vH = _mm_load_si128(pvHStore + iter - 1);
         vH = _mm_slli_si128(vH, 2);
         vH = _mm_or_si128(vH, vMin);

         pv = pvHLoad;
         pvHLoad = pvHStore;
         pvHStore = pv;

         pv = pvEGLLoad;
         pvEGLLoad = pvEGLStore;
         pvEGLStore = pv;

         for (j = 0; j < iter; j++) {
            /* load values of vE from previous row (one unit up) */
            vE = _mm_load_si128(pvE + j);

            /* load query gap lengths */
            vEGL = _mm_load_si128(pvEGLLoad + j);

            /* add score to vH */
            vH = _mm_adds_epi16(vH, *pvScore++);

            /**
             *  All code related to backtracing was added by Oleksiy Kohany 02/24/2009
             *  vEGL - query gap lengths
             *  vFGL - database gap lengths
             */
            /* Let's register the position of a new maximum score */
            vMSK = _mm_cmpgt_epi16(vH, vMaxScore);
            vMSK2 = _mm_set1_epi16(i - SHORT_BIAS);
            vMSK2 = _mm_and_si128(vMSK, vMSK2);
            vMaxScore_i = _mm_andnot_si128(vMSK, vMaxScore_i);
            vMaxScore_i = _mm_or_si128(vMaxScore_i, vMSK2);

            vMSK2 = _mm_set1_epi16(j);
            vMSK2 = _mm_and_si128(vMSK, vMSK2);
            vMaxScore_j = _mm_andnot_si128(vMSK, vMaxScore_j);
            vMaxScore_j = _mm_or_si128(vMaxScore_j, vMSK2);

            /* Update highest score encountered this far */
            vMaxScore = _mm_max_epi16(vMaxScore, vH);

            /* adding query and database gaps to traceback vector */
            vMSK = _mm_cmpgt_epi16(vE, vH); //Mask vE>vH
            vDIR = _mm_and_si128(vEGL, vMSK); //COPY vEGL for vE>vH to vDIR
            vMSK = _mm_cmpgt_epi16(vE, vF); //Mask vE>vF
            vDIR = _mm_and_si128(vDIR, vMSK); //Mask out vDIR for vE<=vF
            vMSK2 = _mm_cmpgt_epi16(vF, vH); //Mask vF>vH
            vMSK2 = _mm_and_si128(vFGL, vMSK2); //Copy vFGL for vF>vH
            vMSK2 = _mm_andnot_si128(vMSK, vMSK2); //Remove vFGL where vF<=vE
            vDIR = _mm_subs_epi16(vDIR, vMSK2); //Copy negative vFGL


            /* get max from vH, vE and vF */
            vH = _mm_max_epi16(vH, vE);
            vH = _mm_max_epi16(vH, vF);

            /* adding traceback termination marks to traceback vector */
            vMSK = _mm_cmpeq_epi16(vH, vMinimums);
            vMSK = _mm_and_si128(vMSK, vMinimums);
            vDIR = _mm_or_si128(vDIR, vMSK);

            /* save vH values */
            _mm_store_si128(pvHStore + j, vH);

            /* creating gap initialization scores */
            vH = _mm_subs_epi16(vH, vGapOpen);

            /* creating gap extension scores */
            vE = _mm_subs_epi16(vE, vGapExtend);
            vF = _mm_subs_epi16(vF, vGapExtend);

            /* updating query gap lengths vector */
            vMSK = _mm_cmpgt_epi16(vE, vH);
            vEGL = _mm_and_si128(vEGL, vMSK);
            vEGL = _mm_add_epi16(vEGL, vONE);

            /* updating db gap lengths vector */
            vMSK = _mm_cmpgt_epi16(vF, vH);
            vFGL = _mm_and_si128(vFGL, vMSK);
            vFGL = _mm_add_epi16(vFGL, vONE);

            /* update vE and vF values */
            vE = _mm_max_epi16(vE, vH);
            vF = _mm_max_epi16(vF, vH);

            /* save vE, query gap lengths values, and tracebacks  */
            _mm_store_si128(pvE + j, vE);
            _mm_store_si128(pvEGLStore + j, vEGL);
            _mm_store_si128(pvDIR + j, vDIR);

            /* load the next h value */
            vH = _mm_load_si128(pvHLoad + j);
         }

         /* reset pointers to the start of the saved data */
         j = 0;
         vH = _mm_load_si128(pvHStore + j);

         /*  the computed vF value is for the given column.  since */
         /*  we are at the end, we need to shift the vF value over */
         /*  to the next column. */
         vF = _mm_slli_si128(vF, 2);
         vF = _mm_or_si128(vF, vMin);
         vFGL = _mm_slli_si128(vFGL, 2);
         vTemp = _mm_subs_epi16(vH, vGapOpen);
         vTemp = _mm_cmpgt_epi16(vF, vTemp);
         cmp = _mm_movemask_epi8(vTemp);
         while (cmp != 0x0000) {
            vE = _mm_load_si128(pvE + j);
            vEGL = _mm_load_si128(pvEGLLoad + j);
            vDIR = _mm_load_si128(pvDIR + j);

            vMSK = _mm_cmpgt_epi16(vF, vH);
            vDIR = _mm_and_si128(vFGL, vMSK);
            vMSK = _mm_andnot_si128(vMSK, *(pvDIR + j));
            vDIR = _mm_subs_epi16(vMSK, vDIR);

            vH = _mm_max_epi16(vH, vF);

            /* save vH values */
            _mm_store_si128(pvHStore + j, vH);

            /*  update vE incase the new vH value would change it */
            vH = _mm_subs_epi16(vH, vGapOpen);

            vMSK = _mm_cmpgt_epi16(vE, vH);
            vEGL = _mm_and_si128(vEGL, vMSK);
            vEGL = _mm_add_epi16(vEGL, vONE);

            vE = _mm_max_epi16(vE, vH);

            _mm_store_si128(pvE + j, vE);
            _mm_store_si128(pvEGLStore + j, vEGL);
            _mm_store_si128(pvDIR + j, vDIR);

            /* update vF value */
            vF = _mm_subs_epi16(vF, vGapExtend);
            vFGL = _mm_add_epi16(vFGL, vONE);

            j++;
            if (j >= iter) {
               j = 0;
               vF = _mm_slli_si128(vF, 2);
               vF = _mm_or_si128(vF, vMin);
               vFGL = _mm_slli_si128(vFGL, 2);
            }

            vH = _mm_load_si128(pvHStore + j);

            vTemp = _mm_subs_epi16(vH, vGapOpen);
            vTemp = _mm_cmpgt_epi16(vF, vTemp);
            cmp = _mm_movemask_epi8(vTemp);
         }
         /*
          * START
          * Filling in trace matrix for alignment recovery
          */
         short *path = (short*) pvDIR;
         int values_per_vector = 8;
         int wi, qi;
         int mi = 1;
         //For each value in 128bit vector
         for (wi = 0; wi < values_per_vector; wi++) {
            //For each 8-word iteration
            for (qi = 0; qi < iter; qi++) {
               int queryPos = wi + qi * values_per_vector;
               if (mi <= topSeqLength) {
                  int index = mi * (botSeqLength + 1) + i + 1;
                  pathMatrix[index] = path[queryPos];
                  mi++;
               }
            }
         }
         /*
          * END
          */
      }

      short *p_iMaxScore_i = (short*) &vMaxScore_i;
      short *p_iMaxScore_j = (short*) &vMaxScore_j;
      short *p_iMaxScore = (short*) &vMaxScore;
      int iMaxScore_i = 0, iMaxScore_j = 0, iMaxScore_r = 0;
      for (i = 0; i < 8; i++) {
         if (iMaxScore < p_iMaxScore[i]) {
            iMaxScore = p_iMaxScore[i];
            iMaxScore_i = p_iMaxScore_i[i];
            iMaxScore_j = p_iMaxScore_j[i];
            iMaxScore_r = i;
         }
      }
      iMaxScore_i += SHORT_BIAS;
      int iMaxX = iMaxScore_i + 1;
      int iMaxY = iMaxScore_r * iter + iMaxScore_j + 1;
      double score = (short) iMaxScore + SHORT_BIAS;
      if (score) {
         traceBack(topSeq, botSeq, pathMatrix, iMaxX, iMaxY, top, bot);
      }
      return score;
   }
}
