/*
 * SWAligner.cpp
 *
 *  Created on: Apr 6, 2010
 *      Author: Oleksiy Kohany <kohany@gmail.com>
 */

#include "SWAligner.h"
#include <memory>
using namespace std;
namespace bio {
   const char
         SWAligner::lightWeightMatrix[] =
               "GAP -10 -2\n\
       A   C   G   K   M   N   R   S   T   W   X   Y\n\
   A   3  -4  -1  -3   0  -1   1  -3  -4   0  -1  -4\n\
   C  -4   3  -4  -3   0  -1  -4   0  -2  -3  -1   1\n\
   G  -2  -4   3   0  -3  -1   1   0  -4  -3  -1  -4\n\
   K  -3  -3   0   0  -3  -1  -2  -2   0  -2  -1  -2\n\
   M   0   0  -3  -3   0  -1  -2  -2  -3  -2  -1  -2\n\
   N  -1  -1  -1  -1  -1  -1  -1  -1  -1  -1  -1  -1\n\
   R   1  -4   1  -2  -2  -1   1  -2  -4  -2  -1  -4\n\
   S  -3   0   0  -2  -2  -1  -2   0  -3  -3  -1  -2\n\
   T  -4  -1  -4   0  -3  -1  -4  -3   3   0  -1   1\n\
   W   0  -3  -3  -2  -2  -1  -2  -3   0   0  -1  -2\n\
   X  -1  -1  -1  -1  -1  -1  -1  -1  -1  -1  -1  -1\n\
   Y  -4   1  -4  -2  -2  -1  -4  -2   1  -2  -1   1";

   SWAligner::SWAligner() :
      scoreMatrix(lightWeightMatrix, false) {
      // TODO Auto-generated constructor stub

   }

   SWAligner::~SWAligner() {
      // TODO Auto-generated destructor stub
   }

   SWAligner::vindex_t& SWAligner::string2vector(const string& s, vindex_t& v) const {
      v.clear();
      v.reserve(s.size());
      for (string::const_iterator sptr = s.begin(); sptr != s.end(); ++sptr) {
         v.push_back(scoreMatrix.getChar2Index()[*sptr]);
      }
      return v;
   }
   string& SWAligner::vector2string(const vindex_t& v, string& s) const {
      s.clear();
      s.reserve(v.size());
      for (vindex_t::const_iterator vptr = v.begin(); vptr != v.end(); ++vptr) {
         s.push_back(scoreMatrix.getIndex2Char()[*vptr]);
      }
      return s;
   }
   double SWAligner::align(BioSubSequence& top, BioSubSequence& bot) const {
      vindex_t vtop;
      vindex_t vbot;
      string2vector(top, vtop);
      string2vector(bot, vbot);
      auto_ptr<SwData> topProfile(getProfile(vtop));
      double score = alignmentImplementation(vtop, vbot, abs(scoreMatrix.getGapIniPenalty()), abs(scoreMatrix.getGapExtPenalty()), topProfile.get(),
                                             top, bot);
      if (score == 0) {
         top.setFrom(0);
         top.setTo(0);
         bot.setFrom(0);
         bot.setTo(0);
      }
      return score;
   }

   /**
    * Alignment reconstruction from Gotoh-type traceback matrix.
    */
   int SWAligner::traceBack(const vindex_t& topSeq, const vindex_t& botSeq, const vtrace_t& pathMat, int iMaxX, int iMaxY,
                            BioSubSequence& top, BioSubSequence& bot) const {
      const size_t matHeight = topSeq.size() + 1;
      const size_t matWidth = botSeq.size() + 1;

#ifdef DEBUG
      string qSeq;
      string dSeq;
      vector2string(topSeq, qSeq);
      vector2string(botSeq, dSeq);
      printTraceMatrix(pathMat, matHeight, matWidth, qSeq.c_str(), dSeq.c_str());
#endif

      top.setFrom(0);
      bot.setFrom(0);
      top.setTo(iMaxY);
      bot.setTo(iMaxX);

      size_t x = iMaxX;
      size_t y = iMaxY;

#ifdef DEBUG
      fprintf(stderr, "START [%d:%d]\n", x, y);
#endif

      string topAln;
      string botAln;

      /**
       * We start reconstructing local alignment from position (iMaxX, iMaxY) in reverse order.
       * Meaning of values in pathMat:
       * -0x8000 - local alignment termination marker;
       * v>0     - gap of length v in top sequence;
       * v<0     - gap of length v in bottom sequence;
       * 0       - diagonal movement
       */
      do {
         size_t i = y * matWidth + x;
         if (pathMat[i] == -0x8000) {
            break;
         }
         if (pathMat[i] == 0) {
            topAln.push_back(scoreMatrix.getIndex2Char()[topSeq[--y]]);
            botAln.push_back(scoreMatrix.getIndex2Char()[botSeq[--x]]);
         } else {
            if (pathMat[i] > 0) {
               short gaplen = pathMat[i];
               while (gaplen-- > 0) {
                  topAln.push_back(scoreMatrix.GAPCHAR);
                  botAln.push_back(scoreMatrix.getIndex2Char()[botSeq[--x]]);
               }
            } else {
               if (pathMat[i] < 0) {
                  short gaplen = -pathMat[i];
                  while (gaplen-- > 0) {
                     topAln.push_back(scoreMatrix.getIndex2Char()[topSeq[--y]]);
                     botAln.push_back(scoreMatrix.GAPCHAR);
                  }
               }
            }
         }
      } while (x > 0 && y > 0);

      top.setFrom(y + 1);
      bot.setFrom(x + 1);
      top.resize(topAln.size());
      bot.resize(topAln.size());
      copy(topAln.rbegin(), topAln.rend(), top.begin());
      copy(botAln.rbegin(), botAln.rend(), bot.begin());
      return topAln.size();
   }

   /**
    * These are for debugging purposes
    * */
   void SWAligner::printAlignmentMatrix(const short* scoreMat, int height, int width, const char* rowHead, const char* colHead) {
      fprintf(stderr, "\n");
      fprintf(stderr, ">Q\n%s\n", rowHead);
      fprintf(stderr, ">D\n%s\n", colHead);

      fprintf(stderr, "%5c%5c", ' ', ' ');
      int x, y;
      for (x = 0; x < width - 1; x++) {
         fprintf(stderr, "%c:%-3d", colHead[x], x + 1);
      }
      fprintf(stderr, "\n");
      for (y = 0; y < height; y++) {
         if (y > 0) {
            fprintf(stderr, "%c:%-3d", rowHead[y - 1], y);
         } else {
            fprintf(stderr, "%5c", ' ');
         }
         for (x = 0; x < width; x++) {

            fprintf(stderr, "%-5d", scoreMat[y * width + x]);
         }
         fprintf(stderr, "\n");
      }
   }

   /**
    * These are for debugging purposes
    * */
   void SWAligner::printTraceMatrix(const short* scoreMat, int height, int width, const char* rowHead, const char* colHead) {
      fprintf(stderr, "\n");
      fprintf(stderr, ">Q\n'%s'\n", rowHead);
      fprintf(stderr, ">D\n'%s'\n", colHead);

      fprintf(stderr, "%5c%5c", ' ', ' ');
      int x, y;
      for (x = 0; x < width - 1; x++) {
         fprintf(stderr, "%c:%-3d", colHead[x], x + 1);
      }
      fprintf(stderr, "\n");
      for (y = 0; y < height; y++) {
         if (y > 0) {
            fprintf(stderr, "%c:%-3d", rowHead[y - 1], y);
         } else {
            fprintf(stderr, "%5c", ' ');
         }
         for (x = 0; x < width; x++) {
            switch (scoreMat[y * width + x]) {
            case 0:
               fprintf(stderr, "%-5c", '\\');
               break;
            case -0x8000:
               fprintf(stderr, "%-5c", 'X');
               break;
            default:
               fprintf(stderr, "%-5d", scoreMat[y * width + x]);
            }
         }
         fprintf(stderr, "\n");
      }
   }
}
